function [L,P1,P2,nopt,D]=qsl(X,Y,n,D)

% [L,P1,P2,nopt,D]=qsl(X,Y,n,D)
% 
% qsl computes the posterior probabilities of two classes at the data 
% points in X. The class labels are stored in the vector Y. n denotes the 
% number of points to be included in the reference set for exhaustive 
% nearest neighbor classifications, and D is the symmetrical matrix of
% pairwise distances of the points in X.
% 
% L returns the log-likelihood ratio of the posterior probabilities, 
% while the probabilities themselves are returned in P1 and P2. Under the 
% optimization option, the optimal reference set size is returned in nopt.
% D returns the pairwise distances matrix.
% 
% [L,P1,P2,nopt,D]=qsl(X,Y) or [L,P1,P2,nopt,D]=qsl([],Y,[],D)    carries 
%     out learning by optimizing the reference set size.
% 
% [L,P1,P2,nopt,D]=qsl(X,Y,[nl nu]) or [L,P1,P2,nopt,D]=qsl([],Y,[nl nu],D)
%     carries out the optimization of n between nl and nu.
% 
% [L,P1,P2,nopt,D]=qsl([],Y,n,D) computes the posterior probabilities for 
%     the given n according to the distances in D. If D is an MxN matrix of 
%     distances between Xtest(Mxd) and X(Nxd), then the probabilities are 
%     computed for the points in Xtest in terms of the points in X and their
%     labels in Y.
% 
% If you use the algorithm, please cite:
% 
% B. Karacali, "Quasi-supervised learning for biomedical data analysis,"
% Pattern Recognition, Vol. 43, No. 10, p. 3674-3682, 2010
% 
% Copyright 2011 Bilge Karacali 
% 
% The qsl package is free software; you can redistribute it and/or 
% modify it under the terms of the GNU General Public License as 
% published by the Free Software Foundation; either version 2 of the 
% License, or any later version. 
% 
% The software package is distributed in 
% the hope that it will be useful, but WITHOUT ANY WARRANTY; without even 
% the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
% PURPOSE. See the GNU General Public License for more details.
% 

CL=unique(Y);
Y1=double(Y==CL(1));
N1=sum(Y==CL(1));
N2=sum(Y==CL(2));
N=N1+N2;

global VERBOSITY;
if (length(VERBOSITY)==0)
    VERBOSITY=0;
end

if (nargin<4)
    % compute the pairwise squared distances ..
    D=zeros(N,N);
    for i=1:N
        D((i+1):N,i)=(sum((X((i+1):N,:)-ones(N-i,1)*X(i,:)).^2,2));
        D(i,(i+1):N)=D((i+1):N,i)';
    end
    global SINGLED;
    if (SINGLED)
        D=single(D);
    end
else
    N=size(D,1);
end
if (sum(isnan(D(:)))>0)
    STR='Problem with the distance matrix detected. Willing to continue at your own risk?';
    fprintf(STR);
    pause;
    fprintf('%c',char(8*ones(1,length(STR))));
end

OPTIMIZE=0;
NUNL=0;
if (nargin<3)
    OPTIMIZE=1;
else
    if (length(n)==0)
        OPTIMIZE=1;
        NUNL=0;
    else
        if (length(n)>1)
            OPTIMIZE=1;
            NUNL=1;
        end
    end
end

P1=zeros(N,1);
P2=zeros(N,1);
if (OPTIMIZE)
    % compute the optimal n ..
    if (NUNL==0)
        nu=floor((N1*N2+N1+1+N2)/(N1+2+N2));
        % nu=min([N1 N2])-1;
        nl=1;
    else
        nu=max(n);
        nl=min(n);
    end
    Hn=zeros(1,nu);
    while (min(Hn(nl:nu))==0)
        ns=round((0:4)*(nu-nl)/4+nl);
        if (VERBOSITY)
            fprintf('%10d ',ns);
            fprintf('\n');
        end
        Hs=zeros(1,length(ns));
        nind=0;
        COND=1;
        while (COND)
            nind=nind+1;
            if (Hn(ns(nind))==0)
                % carry out the QSL algorithm ..
                for i=1:N
                    ii=[1:(i-1) (i+1):N];
                    [nouse,o]=sort(D(i,ii));
                    cY1=Y1(ii(o));
                    P1(i)=qslmodule(cY1,ns(nind));
                end
                P2=1-P1;
                
                Hs(nind)=sum(4*P1.*P2)+2*ns(nind);
                if (sum(Hn>0)==0)
                    minP1=P1;
                    minP2=P2;
                    minn=ns(nind);
                else
                    if (Hs(nind)<min(Hn(find(Hn))))
                        minP1=P1;
                        minP2=P2;
                        minn=ns(nind);
                    end
                end
                Hn(ns(nind))=Hs(nind);
            else
                Hs(nind)=Hn(ns(nind));
            end
            if (VERBOSITY)
                fprintf('%10.2f ',Hs(nind));
            end
            if (nind>1)
                if (Hs(nind)>Hs(nind-1))
                    % no need to go further ..
                    COND=0;
                else
                    COND=(nind<length(ns));
                end
            end
        end
        bestnind=min(find(Hs(1:nind)==min(Hs(1:nind))));
        if (VERBOSITY)
            fprintf('\n');
        end
        if (bestnind==1)
            nl=ns(1);
            nu=ns(2);
        else
            if (bestnind==nind)
                nl=ns(nind-1);
                nu=ns(nind);
            else
                nl=ns(bestnind-1);
                nu=ns(bestnind+1);
            end
        end
    end
    n=ns(bestnind);
    nopt=n;
    P1=minP1;
    P2=minP2;
else
    nopt=NaN;
    % carry out the QSL algorithm ..
    ii=1:size(D,2);
    LEARN=((sum(D(:)-reshape(D',size(D,1)*size(D,2),1)~=0)==0)&(max(abs(diag(D)))==0));
    for i=1:N
        if (LEARN)
            ii=[1:(i-1) (i+1):size(D,2)];
        end
        [nouse,o]=sort(D(i,ii));
        cY1=Y1(ii(o));
        P1(i)=qslmodule(cY1,n);
    end
    P2=1-P1;
end

I=find((P1>0)&(P2>0));
L=NaN*zeros(N,1);
L(I)=log2(P1(I)./P2(I));


