%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% this code provides an example run of the qsl algorithm provided by the
% qsl package ..
% 
% If you use this code, please cite:
% 
% B. Karacali, "Quasi-supervised learning for biomedical data analysis,"
% Pattern Recognition, Vol. 43, No. 10, p. 3674-3682, 2010
% 
% Copyright 2011 Bilge Karacali 
% 
% qsldemo is part of the qsl package.
% 
% The qsl package is free software; you can redistribute it and/or 
% modify it under the terms of the GNU General Public License as 
% published by the Free Software Foundation; either version 2 of the 
% License, or any later version. 
% 
% The software package is distributed in 
% the hope that it will be useful, but WITHOUT ANY WARRANTY; without even 
% the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
% PURPOSE. See the GNU General Public License for more details.
% 

clear;

% the control group C0 will consist of points drawn from the control 
% distribution, a bi-variate zero mean and unit covariance Gaussian.

% the mixed group C1 will consist of points drawn from the control 
% distribution as well as the target distribution, another bi-variate 
% Gaussian centered at the point [2, 0] and with unit covariance.

X0=randn(1000,2);
X1=[randn(750,2); randn(250,2)+repmat([2 0],250,1)];

figure(1)
clf
plot(X0(:,1),X0(:,2),'.');
hold on
plot(X1(:,1),X1(:,2),'r.');
xlabel('first variate');
ylabel('second variate');
legend('x in C_0','x in C_1');
title('Vectors in C_0 and C_1');
axis equal
grid on

% now, apply qsl to compute the posterior probabilities of the control and
% mixed groups at each point in X0 and X1.
X=[X0; X1];
Y=[zeros(1000,1); ones(1000,1)];

global VERBOSITY;
VERBOSITY=1; % to follow the progression

[L,P0,P1,nopt]=qsl(X,Y);

% for a memory-friendly alternative, try
% [L,P0,P1,nopt]=batchqsl(X,Y);

% finally, display the estimated posterior probabilities with respect to 
% the first variate against the true (unknown) values.
I0=find(Y==0);
I1=find(Y==1);

t=(0:.01:1)*range(X(:,1))+min(X(:,1));
P0curve=normpdf(t,0,1)./(normpdf(t,0,1)+(0.75*normpdf(t,0,1)+0.25*normpdf(t,2,1)));
P1curve=(0.75*normpdf(t,0,1)+0.25*normpdf(t,2,1))./(normpdf(t,0,1)+(0.75*normpdf(t,0,1)+0.25*normpdf(t,2,1)));

figure(2)
clf
subplot(2,1,1)
plot(X(I0,1),P0(I0),'.');
hold on
plot(X(I1,1),P0(I1),'r.');
plot(t,P0curve,'g');
title('Pr\{C_0|x\} curves');
legend('x in C_0','x in C_1','theoretical');
subplot(2,1,2)
plot(X(I0,1),P1(I0),'.');
hold on
plot(X(I1,1),P1(I1),'r.');
plot(t,P1curve,'g');
title('Pr\{C_1|x\} curves');
legend('x in C_0','x in C_1','theoretical');

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% For detection of the points in C1 from the target distribution:
% 
% Let PD represent the points in C1 for which P1>=Pc, a decision 
% threshold. Let also PFA represent the points in C0 that satisfy the same
% detection rule.

Pcs=1:-.01:0;
PD=[0 zeros(1,length(Pcs)) 1];
PFA=[0 zeros(1,length(Pcs)) 1];
for i=1:length(Pcs)
    PD(i+1)=mean(P1(I1)>=Pcs(i));
    PFA(i+1)=mean(P1(I0)>=Pcs(i));
end
ii=11:10:61;
figure(3)
clf
plot(PFA,PD)
hold on
plot(PFA(ii+1),PD(ii+1),'k.')
for i=1:length(ii)
    text(PFA(ii(i)+1)+.01,PD(ii(i)+1)-.01,sprintf('P_c = %.2f',Pcs(ii(i))),'FontSize',8);
end
xlabel('P_{FA}');
ylabel('P_D');
title('P_D-P_{FA} curve');
grid on
axis equal
axis([0 1 0 1]);

% the points that are specific to C_1 beyond a statistical significance 
% level of alpha:
alpha=0.05;
inds0=find(P1(I0)>=1-alpha);
inds1=find(P1(I1)>=1-alpha);

figure(4)
clf
plot(X(I0,1),X(I0,2),'.','Color',[.5 .5 1]);
hold on
plot(X(I1,1),X(I1,2),'.','Color',[1 .5 .5]);
if (length(inds0)>0)
    plot(X(I0(inds0),1),X(I0(inds0),2),'o');
    plot(X(I1(inds1),1),X(I1(inds1),2),'ro');
    legend('x in C_0','x in C_1','misdetections','true detections');
else
    plot(X(I1(inds1),1),X(I1(inds1),2),'ro');
    legend('x in C_0','x in C_1','true detections');
end
xlabel('first variate');
ylabel('second variate');
title(sprintf('Detection of points specific to C_1 with statistical significance %.4f',alpha));
axis equal
grid on



